/*
 * Copyright (c) 2010,2011 Red Hat, Inc.
 *
 * This file is part of HekaFS.
 *
 * HekaFS is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License, version 3, as published by the Free
 * Software Foundation.
 *
 * HekaFS is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License *
 * along with HekaFS.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ctype.h>
#include <sys/uio.h>

#ifndef _CONFIG_H
#define _CONFIG_H
#include "config.h"
#endif

#include "glusterfs.h"
#include "xlator.h"
#include "logging.h"

#include "crypt.h"


/*
 * TBD: make this endian-neutral.  Right now it works on homogeneous machines
 * because they'll all do the addition the same way, but it should work on
 * heterogeneous machines as well.
 */
void
increment_iv (unsigned char *iv, unsigned int n)
{
	unsigned int *iv_ints = (unsigned int *)iv;
	unsigned char i;

	/* Check for safety from 32-bit overflow. */
	if (n <= (UINT_MAX - iv_ints[3])) {
		iv_ints[3] += n;	
		return;
	}

	/*
	 * Do this carefully to avoid actually hitting the overflow.  We know
	 * that iv_ints[3] cannot be zero because then n could not have been
	 * greater than the remainder and we wouldn't be here.
	 */
	iv_ints[3] = n - (UINT_MAX - iv_ints[3] + 1);

	/* Propagate the carry bit. */
	i = 2;
	for (;;) {
		if (iv_ints[i] != UINT_MAX) {
			/* Carry bit stops here. */
			++iv_ints[i];
			break;
		}
		iv_ints[i] = 0;
		if (i == 0) {
			/* Total overflow; wraparound is OK. */
			break;
		}
		/* Propagate (at least) one further. */
		--i;
	}
}

void
encrypt_chunk (AES_KEY *key, unsigned char *input, unsigned char *output,
	       unsigned char *gfid, off_t file_offset, size_t length)
{
	size_t        length_now;	/* length within data block */
	unsigned char iv_input[AES_BLOCK_SIZE];
	unsigned char actual_iv[AES_BLOCK_SIZE];
	unsigned char keystream[AES_BLOCK_SIZE];
	unsigned char ib_offset;	/* intra-block offset */

	while (length > 0) {
		length_now = BLOCK_SIZE - (file_offset % BLOCK_SIZE);
		if (length_now > length) {
			length_now = length;
		}
		/* Generate a hard-to-predict IV for the block, using a method
		 * similar to ESSIV but without the redundant hash step.  Note
		 * that the IV will still be constant for the same key, GFID,
		 * and block number.  This does make us susceptible to a
		 * known-plaintext attack; anyone who can see both plaintext
		 * and ciphertext for a block can derive that section of the
		 * keystream even without having the key, and can then use that
		 * to read any plaintext in that same block thereafter.  So
		 * don't let anyone see both.  Really.  If you send plaintext
		 * to the same people who receive the ciphertext (i.e. the
		 * people who run your storage servers) then you'll need to
		 * change either the key or the GFID to regain confidentiality.
		 * This is simple if you just do copy/rename, but involves
		 * a lot of data movement and calculation.
		 */
		memset(iv_input,0,sizeof(iv_input));
		memcpy(iv_input,gfid,GFID_SIZE);
		*((off_t *)iv_input) = file_offset / BLOCK_SIZE;
		AES_encrypt(iv_input,actual_iv,key);
		increment_iv(actual_iv,(file_offset%BLOCK_SIZE)/AES_BLOCK_SIZE);
		/*
		 * Now for the actual encryption.  This is very similar to
		 * AES_ctr128_encrypt, but adjusted to avoid API inefficiency
		 * and to use iv_increment for consistency with the above.
		 */
		AES_encrypt(actual_iv,keystream,key);
		DPRINTF("keystream for %"PRIu64":%"PRIu64" starts with %02x %02x %02x\n",
			file_offset / BLOCK_SIZE,
			(file_offset % BLOCK_SIZE) / AES_BLOCK_SIZE,
			keystream[0], keystream[1], keystream[2]);
		ib_offset = file_offset % AES_BLOCK_SIZE;
		for (;;) {
			*(output++) = *(input++) ^ keystream[ib_offset];
			++file_offset;
			--length;
			if (--length_now == 0) {
				break;
			}
			if (++ib_offset == AES_BLOCK_SIZE) {
				increment_iv(actual_iv,1);
				AES_encrypt(actual_iv,keystream,key);
				ib_offset = 0;
			}
		}
	}
}


int32_t
crypt_readv_cbk (call_frame_t *frame,
                 void *cookie,
                 xlator_t *this,
                 int32_t op_ret,
                 int32_t op_errno,
                 struct iovec *vector,
                 int32_t count,
		 struct iatt *stbuf,
                 struct iobref *iobref)
{
	crypt_rlocal_t  *local = frame->local;
	crypt_private_t *priv = this->private;
	int32_t          i = 0;
	off_t            cur_off = local->offset;

	for (i = 0; i < count; ++i) {
		encrypt_chunk(&priv->key,vector[i].iov_base,vector[i].iov_base,
			      local->gfid,cur_off,vector[i].iov_len);
		cur_off += vector[i].iov_len;
	}

	STACK_UNWIND_STRICT(readv,frame,op_ret,op_errno,
			    vector,count,stbuf,iobref);
	return 0;
}

/*
 * We can only decrypt starting at block boundaries, so we might need to read
 * a little extra at the beginning if this read is unaligned.  Otherwise, we
 * could just pass this through and do all our work in the callback.
 */
int32_t
crypt_readv (call_frame_t *frame,
             xlator_t *this,
             fd_t *fd,
             size_t size,
             off_t offset)
{
	crypt_rlocal_t  *local = NULL;
	uint32_t         op_errno = EIO;

	local = CALLOC(1,sizeof(*local));
	if (!local) {
		op_errno = ENOMEM;
		goto err;
	}
	frame->local = local;

	local->offset = offset;
	memcpy(local->gfid,fd->inode->gfid,sizeof(local->gfid));

	STACK_WIND (frame,
		    crypt_readv_cbk,
		    FIRST_CHILD (this),
		    FIRST_CHILD (this)->fops->readv,
		    fd, size, offset);
	return 0;

err:
	STACK_UNWIND_STRICT(readv,frame,-1,op_errno,NULL,0,NULL,NULL);
	return 0;
}


int32_t
crypt_writev_cbk (call_frame_t *frame,
                  void *cookie,
                  xlator_t *this,
                  int32_t op_ret,
                  int32_t op_errno,
                  struct iatt *prebuf,
		  struct iatt *postbuf)
{

	STACK_UNWIND_STRICT (writev, frame, op_ret, op_errno, prebuf, postbuf);
	return 0;
}


int32_t
crypt_writev (call_frame_t *frame,
              xlator_t *this,
              fd_t *fd,
              struct iovec *vector,
              int32_t count,
              off_t offset,
              struct iobref *iobref)
{
	crypt_private_t *priv = this->private;
	int32_t          i = 0;
	off_t            cur_off = offset;

	for (i = 0; i < count; ++i) {
		encrypt_chunk(&priv->key,vector[i].iov_base,vector[i].iov_base,
			      fd->inode->gfid,cur_off,vector[i].iov_len);
		cur_off += vector[i].iov_len;
	}

	STACK_WIND (frame, crypt_writev_cbk,
		    FIRST_CHILD (this), FIRST_CHILD (this)->fops->writev,
		    fd, vector, count, offset, iobref);
	return 0;
}

int32_t
crypt_set_key (data_t *data, AES_KEY *key)
{
	int           rc = -1;
	unsigned char hex_buf[32] = {0};	/* binary AES-256 */
	unsigned char i = 0;
	int           hex_byte = 0;
	int           fd = -1;
	unsigned char file_buf[64] = {0};	/* hex AES-256 */

	if (!data) {
		gf_log(__func__,GF_LOG_ERROR,"missing key option");
		return EINVAL;
	}
	gf_log(__func__,GF_LOG_DEBUG,"data length is %d",data->len);

	/*
	 * Mostly we let AES_set_encrypt_key do key-size checking, since
	 * it'll do it anyway even if we already did.  The exceptions are (1)
	 * a null key which is practically free to check in the switch, and
	 * (2) the hex-key length which could cause an overflow in this code
	 * before we ever call AES_set_encrypt_key.
	 */
	switch (data->data[0]) {
	case '\0':
		gf_log(__func__,GF_LOG_DEBUG,"null key");
		return EINVAL;
	case '%':	/* hex key */
		gf_log(__func__,GF_LOG_DEBUG,"handling hex key");
		switch (data->len) {
		case 34:	/* AES-128 = "%<32x>\0" */
		case 50:	/* AES-192 = "%<48x>\0" */
		case 66:	/* AES-256 = "%<64x>\0" */
			break;
		default:
			gf_log(__func__,GF_LOG_DEBUG,"bad hex-key length");
			return EINVAL;
		}
		for (i = 0; i < (data->len / 2 - 1); ++i) {
			if (sscanf(data->data+i*2+1,"%2x",&hex_byte) != 1) {
				break;
			}
			hex_buf[i] = hex_byte & 0xff;
		}
		rc = AES_set_encrypt_key(hex_buf,i*8,key);
		break;
	case '/':	/* key in file */
		gf_log(__func__,GF_LOG_DEBUG,"handling file key");
		fd = open(data->data,O_RDONLY);
		if (fd < 0) {
			gf_log(__func__,GF_LOG_ERROR,"could not open key file");
			return EINVAL;
		}
		rc = read(fd,file_buf,sizeof(file_buf));
		close(fd);
		for (i = 0; i < (rc / 2); ++i) {
			if (sscanf(file_buf+i*2,"%2x",&hex_byte) != 1) {
				break;
			}
			hex_buf[i] = hex_byte & 0xff;
		}
		rc = AES_set_encrypt_key(hex_buf,i*8,key);
		break;
	default:	/* text key */
		gf_log(__func__,GF_LOG_DEBUG,"handling text key");
		rc = AES_set_encrypt_key(data->data,(data->len-1)*8,key);
	}

	return rc ? EINVAL : 0;
}

int32_t
init (xlator_t *this)
{
	crypt_private_t *priv = NULL;
	int32_t          status = 0;

	if (!this->children || this->children->next) {
		gf_log ("crypt", GF_LOG_ERROR,
			"FATAL: crypt should have exactly one child");
		return -1;
	}

	if (!this->parents) {
		gf_log (this->name, GF_LOG_WARNING,
			"dangling volume. check volfile ");
	}

	priv = CALLOC (sizeof (crypt_private_t), 1);
	if (!priv) {
		return -1;
	}
	this->private = priv;

	/*
	 * TBD: the option should really specify a file containing a longer key
	 * for a better encryption algorithm.
	 */
	status = crypt_set_key(dict_get(this->options,"key"),&priv->key);
	if (status != 0) {
		gf_log(this->name,GF_LOG_ERROR,"key missing");
		return status;
	}
	gf_log ("crypt", GF_LOG_INFO, "crypt xlator loaded");
	return 0;
}

void
fini (xlator_t *this)
{
	crypt_private_t *priv = this->private;

	FREE (priv);

	return;
}

struct xlator_fops fops = {
	.readv        = crypt_readv,
	.writev       = crypt_writev,
};

struct xlator_cbks cbks = {
};

struct volume_options options[] = {
	{ .key = {"key"},
	  .type = GF_OPTION_TYPE_STR
	},
	{ .key  = {NULL} },
};
