/*
 * Copyright 2011 Red Hat, Inc.
 *
 * This file is part of HekaFS.
 *
 * HekaFS is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License, version 3, as published by the Free
 * Software Foundation.
 *
 * HekaFS is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License *
 * along with HekaFS.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/file.h>
#include <stdio.h>
#include "glusterfs.h"
#include "xlator.h"
#include "uidmap.h"

typedef struct {
        char *name;
        uid_t uid_low;
        uid_t uid_high;
        uid_t uid_next;
        gid_t gid_low;
        gid_t gid_high;
        uid_t gid_next;
        struct rb_table *uid_map;
        struct rb_table *uid_revmap;
        struct rb_table *gid_map;
        struct rb_table *gid_revmap;
        int root_squash;
        pthread_mutex_t mtx;
} mapping_t;

union tuple {
        struct {
                uid_t u1;
                uid_t u2;
        };
        struct {
                gid_t g1;
                gid_t g2;
        };
};

#define RBMAP_SIGNATURE         "hekafs map: rbmap\n"
#define RBMAP_UID_MAP_ENTRY     "uid_map_entry:"
#define RBMAP_UID_MAP_ENTRY_LEN 13
#define RBMAP_UID_LOW           "uid_low:"
#define RBMAP_UID_LOW_LEN       8
#define RBMAP_UID_HIGH          "uid_high:"
#define RBMAP_UID_HIGH_LEN      9
#define RBMAP_UID_NEXT          "uid_next:"
#define RBMAP_UID_NEXT_LEN      9
#define RBMAP_GID_MAP_ENTRY     "gid_map_entry:"
#define RBMAP_GID_MAP_ENTRY_LEN 13
#define RBMAP_GID_LOW           "gid_low:"
#define RBMAP_GID_LOW_LEN       8
#define RBMAP_GID_HIGH          "gid_high:"
#define RBMAP_GID_HIGH_LEN      9
#define RBMAP_GID_NEXT          "gid_next:"
#define RBMAP_GID_NEXT_LEN      9

#define RBMAP_LOG_LEVEL GF_LOG_DEBUG

static struct rb_table *mappings;
static pthread_mutex_t mappings_mtx = PTHREAD_MUTEX_INITIALIZER;

/* rb decls so we don't have to carry Gluster's rb.h */
#define RB_MAX_HEIGHT 48
struct rb_traverser {
	struct rb_table *rb_table;
	struct rb_node *rb_node;
	struct rb_node *rb_stack[RB_MAX_HEIGHT];
	size_t rb_height;
	unsigned long rb_generation;
};

extern void *
rb_t_first(struct rb_traverser *, struct rb_table *);

extern void *
rb_t_next(struct rb_traverser *);

extern void*
rb_insert(struct rb_table *, void *);

extern void*
rb_find(const struct rb_table *, const void *);

extern void **
rb_probe(struct rb_table *, void *);

extern void *
rb_delete(struct rb_table *, const void *);

typedef int rb_comparison_func(const void *, const void *, void *);

extern struct rb_table *
rb_create(rb_comparison_func *, void *, void *);

typedef void rb_item_func (void *, void *);

extern void
rb_destroy(struct rb_table *, rb_item_func *);

/* serialize and persist */
static void
rbmap_serialize(mapping_t *mapping)
{
        char fname[128], lkname[128];
        FILE *file = NULL;
        int lockfd = -1;
        (void) snprintf(lkname, sizeof lkname, UIDMAP_MAP_LOCK_FILE, mapping->name);
        (void) snprintf(fname, sizeof fname, UIDMAP_MAP_FILE, mapping->name);

        do {
                int status;
                struct rb_traverser trav;
                union tuple *data;
                lockfd = open(lkname, O_CREAT|O_WRONLY, 0644);
                if (lockfd == -1) {
                        gf_log(mapping->name, GF_LOG_CRITICAL,
                               "open %s: %s", lkname, strerror(errno));
                        break;
                }
                status = flock(lockfd, LOCK_EX);
                if (status == -1) {
                        gf_log(mapping->name, GF_LOG_CRITICAL,
                               "flock failed %s", strerror(errno));
                        break;
                }
                file = fopen(fname, "w");
                if (file == NULL) {
                        (void) flock(lockfd, LOCK_UN);
                        gf_log(mapping->name, GF_LOG_CRITICAL,
                               "fopen failed %s", strerror(errno));
                        break;
                }
                (void) fprintf(file, RBMAP_SIGNATURE);
                data = (union tuple *) rb_t_first(&trav, mapping->uid_map);
                while (data != NULL) {
                        (void) fprintf(file, "%s %u %u\n",
                                       RBMAP_UID_MAP_ENTRY,
                                       data->u1, data->u2);
                        data = (union tuple *) rb_t_next(&trav);
                }
                (void) fprintf(file, "%s %u\n", RBMAP_UID_LOW, mapping->uid_low);
                (void) fprintf(file, "%s %u\n", RBMAP_UID_HIGH, mapping->uid_high);
                (void) fprintf(file, "%s %u\n", RBMAP_UID_NEXT, mapping->uid_next);
                data = (union tuple *) rb_t_first(&trav, mapping->gid_map);
                while (data != NULL) {
                        (void) fprintf(file, "%s %u %u\n",
                                       RBMAP_GID_MAP_ENTRY,
                                       data->g1, data->g2);
                        data = (union tuple *) rb_t_next(&trav);
                }
                (void) fprintf(file, "%s %u\n", RBMAP_GID_LOW, mapping->gid_low);
                (void) fprintf(file, "%s %u\n", RBMAP_GID_HIGH, mapping->gid_high);
                (void) fprintf(file, "%s %u\n", RBMAP_GID_NEXT, mapping->gid_next);
        } while (0);
        if (file != NULL) {
                (void) fclose(file);
        }
        if (lockfd != -1) {
                (void) flock(lockfd, LOCK_UN);
                (void) close(lockfd);
        }
}


static int
rbmap_readmapfile(mapping_t *mapping, FILE *file)
{
        char scratch[128];
        char* sts;
        while ((sts = fgets(scratch, sizeof scratch, file)) != (char *) EOF && sts != NULL) {
                if (strncmp(scratch, RBMAP_UID_MAP_ENTRY, RBMAP_UID_MAP_ENTRY_LEN) == 0) {
                        union tuple *map_data;
                        union tuple *revmap_data =
                                (union tuple *) malloc(sizeof(union tuple));
                        if (revmap_data != NULL) {
                                map_data = (union tuple *) malloc(sizeof(union tuple));
                        } else {
                                return -1;
                        }
                        if (map_data != NULL) {
                                int num = sscanf(&scratch[RBMAP_UID_MAP_ENTRY_LEN+1],
                                                 " %u %u", &map_data->u1, &map_data->u2);
                                if (num == 2) {
                                        rb_insert(mapping->uid_map, map_data);
                                        revmap_data->u1 = map_data->u2;
                                        revmap_data->u2 = map_data->u1;
                                        rb_insert(mapping->uid_revmap, revmap_data);
                                } else {
                                        (void) free(map_data);
                                        (void) free(revmap_data);
                                        return -1;
                                }
                        } else {
                                (void) free(revmap_data);
                                return -1;
                        }
                } else if (strncmp(scratch, RBMAP_UID_LOW, RBMAP_UID_LOW_LEN) == 0) {
                        mapping->uid_low =
                                (uid_t) strtoul(&scratch[RBMAP_UID_LOW_LEN+1], NULL, 10);
                } else if (strncmp(scratch, RBMAP_UID_HIGH, RBMAP_UID_HIGH_LEN) == 0) {
                        mapping->uid_high =
                                (uid_t) strtoul(&scratch[RBMAP_UID_HIGH_LEN+1], NULL, 10);
                } else if (strncmp(scratch, RBMAP_UID_NEXT, RBMAP_UID_NEXT_LEN) == 0) {
                        mapping->uid_next =
                                (uid_t) strtoul(&scratch[RBMAP_UID_NEXT_LEN+1], NULL, 10);
                } else if (strncmp(scratch, RBMAP_GID_MAP_ENTRY, RBMAP_GID_MAP_ENTRY_LEN) == 0) {
                        union tuple *map_data;
                        union tuple *revmap_data =
                                (union tuple *) malloc(sizeof(union tuple));
                        if (revmap_data != NULL) {
                                map_data = (union tuple *) malloc(sizeof(union tuple));
                        } else {
                                return -1;
                        }
                        if (map_data != NULL) {
                                int num = sscanf(&scratch[RBMAP_GID_MAP_ENTRY_LEN+1],
                                                 " %u %u", &map_data->g1, &map_data->g2);
                                if (num == 2) {
                                        rb_insert(mapping->gid_map, map_data);
                                        revmap_data->g1 = map_data->g2;
                                        revmap_data->g2 = map_data->g1;
                                        rb_insert(mapping->gid_revmap, revmap_data);
                                } else {
                                        (void) free(map_data);
                                        (void) free(revmap_data);
                                        return -1;
                                }
                        } else {
                                (void) free(revmap_data);
                                return -1;
                        }
                } else if (strncmp(scratch, RBMAP_GID_LOW, RBMAP_GID_LOW_LEN) == 0) {
                        mapping->gid_low =
                                (uid_t) strtoul(&scratch[RBMAP_GID_LOW_LEN+1], NULL, 10);
                } else if (strncmp(scratch, RBMAP_GID_HIGH, RBMAP_GID_HIGH_LEN) == 0) {
                        mapping->gid_high =
                                (uid_t) strtoul(&scratch[RBMAP_GID_HIGH_LEN+1], NULL, 10);
                } else if (strncmp(scratch, RBMAP_GID_NEXT, RBMAP_GID_NEXT_LEN) == 0) {
                        mapping->gid_next =
                                (uid_t) strtoul(&scratch[RBMAP_GID_NEXT_LEN+1], NULL, 10);
                }
        }
        return 0;
}

static int
rbmap_deserialize(mapping_t *mapping)
{
        char fname[128], lkname[128];
        FILE *file = NULL;
        int lockfd = -1;
        (void) snprintf(lkname, sizeof lkname, UIDMAP_MAP_LOCK_FILE, mapping->name);
        (void) snprintf(fname, sizeof fname, UIDMAP_MAP_FILE, mapping->name);

        do {
                char scratch[128];
                char* sts;
                int status;
                lockfd = open(lkname, O_CREAT|O_RDONLY, 0644);
                if (lockfd == -1) {
                        gf_log(mapping->name, GF_LOG_CRITICAL,
                               "open %s: %s", lkname, strerror(errno));
                        break;
                }
                status = flock(lockfd, LOCK_EX);
                if (status == -1) {
                        gf_log(mapping->name, GF_LOG_CRITICAL,
                               "flock: %s", strerror(errno));
                        break;
                }
                file = fopen(fname, "r");
                if (file == NULL) {
                        gf_log(mapping->name, GF_LOG_INFO,
                               "fopen failed: %s", strerror(errno));
                        break;
                }
                /* read the first line (signature) of the file) */
                if ((sts = fgets(scratch, sizeof scratch, file)) == (char *) EOF || sts == NULL) {
                        gf_log(mapping->name, GF_LOG_CRITICAL,
                               "fgets signature failed: %s", strerror(errno));
                        break;
                }
                if (strcmp(RBMAP_SIGNATURE, scratch) != 0) {
                        gf_log(mapping->name, GF_LOG_CRITICAL,
                               "signature mismatch: %s", scratch);
                        break;
                }
                if (rbmap_readmapfile(mapping, file) == -1) {
                        return -1;
                }
        } while (0);
        if (file != NULL) {
                (void) fclose(file);
        }
        if (lockfd != -1) {
                (void) flock(lockfd, LOCK_UN);
                (void) close(lockfd);
        }
        return 0;
}


int
map(struct _call_stack_t *stack, char *name)
{
        int do_persist = 0;
        mapping_t *mapping, mneedle;
        mneedle.name = name;
        mapping = (mapping_t *) rb_find(mappings, &mneedle);
        /* assert(mapping != NULL); */
        if (mapping == NULL) {
                return -1;
        }
        do {
                pthread_mutex_lock(&mapping->mtx);
		if (stack->uid == 0 && mapping->root_squash) {
			stack->uid = HEKAFS_NOBODY;
		}
                else {
                        /* look for an existing match */
                        union tuple *new_map, tneedle;
                        tneedle.u1 = stack->uid;
                        new_map = (union tuple *) rb_find(mapping->uid_map, &tneedle);
                        if (new_map == NULL) {
                                union tuple *new_revmap = NULL;
                                /* couldn't find one, make a new one */
                                if (mapping->uid_next > mapping->uid_high) {
                                        gf_log(name, GF_LOG_CRITICAL,
                                               "no uids in range %u-%u",
                                               mapping->uid_low, mapping->uid_high);
                                        pthread_mutex_unlock(&mapping->mtx);
                                        return -1;
                                }
                                new_map = (union tuple *) malloc(sizeof(union tuple));
                                new_revmap =
                                        (union tuple *) malloc(sizeof(union tuple));
                                if (new_map == NULL || new_revmap == NULL) {
                                        pthread_mutex_unlock(&mapping->mtx);
                                        if (new_map != NULL) (void) free(new_map);
                                        if (new_revmap != NULL) (void) free(new_revmap);
                                        return -1;
                                }
                                do_persist = 1;
                                new_map->u1 = stack->uid;
                                new_map->u2 = mapping->uid_next++;
                                new_revmap->u1 = new_map->u2;
                                new_revmap->u2 = new_map->u1;
                                if (rb_probe(mapping->uid_map, new_map) == NULL) {
                                        pthread_mutex_unlock(&mapping->mtx);
                                        (void) free(new_map);
                                        (void) free(new_revmap);
                                        return -1;
                                }
                                if (rb_probe(mapping->uid_revmap, new_revmap) == NULL) {
                                        pthread_mutex_unlock(&mapping->mtx);
                                        (void) rb_delete(mapping->uid_map, new_map);
                                        (void) free(new_map);
                                        (void) free(new_revmap);
                                        return -1;
                                }
                                gf_log(name, RBMAP_LOG_LEVEL,
                                       "added new uid mapping %u -> %u",
                                       stack->uid, new_map->u2);
                                if (mapping->uid_next > mapping->uid_high) {
                                        gf_log(name, GF_LOG_CRITICAL,
                                               "exhausted uids in range %u-%u",
                                               mapping->uid_low, mapping->uid_high);
                                }
                        }
                        stack->uid = new_map->u2;
                }
		if (stack->gid == 0 && mapping->root_squash) {
			stack->gid = HEKAFS_NOBODY;
		}
                else {
                        /* look for an existing match */
                        union tuple *new_map, tneedle;
                        tneedle.g1 = stack->gid;
                        new_map = (union tuple *) rb_find(mapping->gid_map, &tneedle);
                        if (new_map == NULL) {
                                union tuple *new_revmap = NULL;
                                /* couldn't find one, make a new one */
                                if (mapping->gid_next > mapping->gid_high) {
                                        gf_log(name, GF_LOG_CRITICAL,
                                               "no gids in range %u-%u",
                                               mapping->gid_low, mapping->gid_high);
                                        pthread_mutex_unlock(&mapping->mtx);
                                        return -1;
                                }
                                new_map = (union tuple *) malloc(sizeof(union tuple));
                                new_revmap =
                                        (union tuple *) malloc(sizeof(union tuple));
                                if (new_map == NULL || new_revmap == NULL) {
                                        pthread_mutex_unlock(&mapping->mtx);
                                        if (new_map != NULL) (void) free(new_map);
                                        if (new_revmap != NULL) (void) free(new_revmap);
                                        return -1;
                                }
                                do_persist = 1;
                                new_map->g1 = stack->gid;
                                new_map->g2 = mapping->gid_next++;
                                new_revmap->g1 = new_map->g2;
                                new_revmap->g2 = new_map->g1;
                                if (rb_probe(mapping->gid_map, new_map) == NULL) {
                                        pthread_mutex_unlock(&mapping->mtx);
                                        (void) free(new_map);
                                        (void) free(new_revmap);
                                        return -1;
                                }
                                if (rb_probe(mapping->gid_revmap, new_revmap) == NULL) {
                                        pthread_mutex_unlock(&mapping->mtx);
                                        (void) rb_delete(mapping->gid_map, new_map);
                                        (void) free(new_map);
                                        (void) free(new_revmap);
                                        return -1;
                                }
                                gf_log(name, RBMAP_LOG_LEVEL,
                                       "added new gid mapping %u -> %u",
                                       stack->gid, new_map->g2);
                                if (mapping->gid_next > mapping->gid_high) {
                                        gf_log(name, GF_LOG_CRITICAL,
                                               "exhausted gids in range %u-%u",
                                               mapping->gid_low, mapping->gid_high);
                                }
                        }
                        stack->gid = new_map->g2;
                }
                if (do_persist > 0)
                        rbmap_serialize(mapping);
                pthread_mutex_unlock(&mapping->mtx);
        } while (0);

        return 0;
}


void
revmap(uid_t *uid, gid_t *gid, char *name)
{
        union tuple *tuple, tneedle;
        mapping_t *mapping, mneedle;
        mneedle.name = name;
        mapping = (mapping_t *) rb_find(mappings, &mneedle);
        /* assert(mapping != NULL); */
        if (mapping == NULL) {
                return;
        }
        switch (*uid) {
        case 0:
                /* break; */
        case HEKAFS_NOBODY:
                *uid = 0;
                break;
        default:
                tneedle.u1 = *uid;
                tuple = (union tuple *) rb_find(mapping->uid_revmap, &tneedle);
                if (tuple != NULL) {
                        *uid = tuple->u2;
                }
                break;
        }
        switch (*gid) {
        case 0:
                /* break; */
        case HEKAFS_NOBODY:
                *uid = 0;
                break;
        default:
                tneedle.g1 = *gid;
                tuple = (union tuple *) rb_find(mapping->gid_revmap, &tneedle);
                if (tuple != NULL) {
                        *gid = tuple->g2;
                }
                break;
        }
}


static int
rbmap_compare_uid(const void *rb_a, const void *rb_b, void *rb_param)
{
        union tuple *a = (union tuple *)rb_a;
        union tuple *b = (union tuple *)rb_b;
        int ret = 0;
        if (a->u1 < b->u1)
                ret--;
        else if (a->u1 > b->u1)
                ret++;
        return ret;
}


static int
rbmap_compare_gid(const void *rb_a, const void *rb_b, void *rb_param)
{
        union tuple *a = (union tuple *)rb_a;
        union tuple *b = (union tuple *)rb_b;
        int ret = 0;
        if (a->g1 < b->g1)
                ret--;
        else if (a->g1 > b->g1)
                ret++;
        return ret;
}

static int
rbmap_compare_tenantname(const void *rb_a, const void *rb_b, void *rb_param)
{
        mapping_t *needle = (mapping_t *) rb_a;
        mapping_t *haystack = (mapping_t *) rb_b;
        return strcmp(haystack->name, needle->name);
}


int32_t
init(struct _xlator *xlator,
     uid_t uid_low, uid_t uid_high, gid_t gid_low, gid_t gid_high,
     int root_squash)
{
        mapping_t *newmapping;
        int ret = 0;
        pthread_mutex_lock(&mappings_mtx);
        if (mappings == NULL) {
                mappings = rb_create(rbmap_compare_tenantname, NULL, NULL);
        }
        pthread_mutex_unlock(&mappings_mtx);
        if (mappings == NULL) {
                return -1;
        }
        newmapping = (mapping_t *) malloc(sizeof(mapping_t));
        if (newmapping == NULL) {
                return -1;
        }
        pthread_mutex_init(&newmapping->mtx, NULL);
        newmapping->name        = strdup(xlator->name);
        newmapping->root_squash = root_squash;
        newmapping->uid_low     = newmapping->uid_next = uid_low;
        newmapping->uid_high    = uid_high;
        newmapping->gid_low     = newmapping->gid_next = gid_low;
        newmapping->gid_high    = gid_high;
        newmapping->uid_map     = rb_create(rbmap_compare_uid, NULL, NULL);
        newmapping->uid_revmap  = rb_create(rbmap_compare_uid, NULL, NULL);
        newmapping->gid_map     = rb_create(rbmap_compare_gid, NULL, NULL);
        newmapping->gid_revmap  = rb_create(rbmap_compare_gid, NULL, NULL);
        if ((ret = rbmap_deserialize(newmapping)) != -1) {
                pthread_mutex_lock(&mappings_mtx);
                rb_insert(mappings, newmapping);
                pthread_mutex_unlock(&mappings_mtx);
        }
        return ret;
}

static void
delete_mapentry(void *map_entry, void *param)
{
	free(map_entry);
}

static void
delete_mapping(void *mapping, void *param)
{
	mapping_t *delete = (mapping_t *)mapping;
	rb_destroy(delete->uid_map, delete_mapentry);
	rb_destroy(delete->uid_revmap, delete_mapentry);
	rb_destroy(delete->gid_map, delete_mapentry);
	rb_destroy(delete->gid_revmap, delete_mapentry);
}

void
fini(struct _xlator *xlator)
{
	struct rb_table *delete = mappings;
	pthread_mutex_lock(&mappings_mtx);
	mappings = NULL;
	pthread_mutex_unlock(&mappings_mtx);
	rb_destroy(delete, delete_mapping);
}

/* convenience symbols for debugging. */
/* Why weak? Dunno, just because I guess; I don't think it matters  */
extern int
_rb_map(struct _call_stack_t *, struct _xlator *)
        __attribute__ ((weak, alias ("map")));

extern void
_rb_revmap(uid_t *, gid_t *)
        __attribute__ ((weak, alias ("revmap")));

extern int32_t
_rb_init(struct _xlator *, char *, uid_t, uid_t, gid_t, gid_t)
        __attribute__ ((weak, alias ("init")));

extern void
_rb_fini(struct _xlator *)
        __attribute__ ((weak, alias ("fini")));
